import numpy as np
import time
from scipy.ndimage.morphology import distance_transform_edt
import lion.utils.ksp as ut_ksp
import logging

logger = logging.getLogger(__name__)


class KSP:

    def __init__(self, graph):
        self.graph = graph
        try:
            test = self.graph.dists_ba.shape  # noqa
            test = self.graph.preds_ba.shape  # noqa
        except AttributeError:
            raise RuntimeError(
                "Cannot initialize KSP object with a graph without"
                "shortest path trees in both directions!"
            )

    def compute_min_node_dists(self):
        """
        Eppstein's algorithm: Sum up the two SP treest and iterate
        """
        # sum both dists_ab and dists_ba, inst and edges are counted twice!
        double_counted = self.graph.edge_cost.copy()
        # must set to zero in order to avoid inf - inf
        double_counted[double_counted == np.inf] = 0
        summed_dists = (
            self.graph.dists + self.graph.dists_ba - double_counted
        )
        # mins along outgoing edges
        min_shift_dists = np.argmin(summed_dists, axis=1)

        # project back to 2D:
        min_dists_2d = np.zeros(self.graph.instance.shape) + np.inf
        min_shifts_2d = np.zeros(self.graph.instance.shape)
        x_inds = self.graph.stack_array[:, 0]
        y_inds = self.graph.stack_array[:, 1]
        min_shifts_2d[x_inds,
                      y_inds] = min_shift_dists[self.graph.pos2node[x_inds,
                                                                    y_inds]]
        min_dists_2d[x_inds,
                     y_inds] = summed_dists[self.graph.pos2node[x_inds,
                                                                y_inds],
                                            min_shift_dists]
        return min_dists_2d, min_shifts_2d

    def ksp(self, k, thresh=20, cost_add=np.inf):
        """
        Fast KSP method as tradeoff between diversity and cost
        (add additional cost to the paths found so far)
        Arguments:
            self.graph.start_inds, self.graph.dest_inds, k: see other methods
            min_dist: minimum distance from the previously computed paths
                (or distance in which we add a penalty)
            cost_add: cost_add of 0.05 means that 5% of the best path costs is
                the maximum costs that are added
        Returns:
            List of ksp with costs
        """
        tic = time.time()
        best_paths = [self.graph.best_path]

        # get Eppstein distances
        (min_node_dists, min_shift_dists) = self.compute_min_node_dists()

        # make auxiliary array for distance transform
        aux_arr = np.ones(min_node_dists.shape)
        for (x, y) in best_paths[0]:
            aux_arr[x, y] = 0

        _, arr_len = min_node_dists.shape
        for _ in range(k - 1):
            # use distance transform to compute the distances to the paths
            distance_transform = distance_transform_edt(aux_arr)
            corridor = thresh - distance_transform
            corridor[corridor < 0] = 0

            # add penalty (or inf to exclude regions)
            corridor[corridor > 0
                     ] = cost_add * corridor[corridor > 0] / np.max(corridor)
            feasible_vertices = (corridor + 1) * min_node_dists

            if ~np.any(feasible_vertices < np.inf):
                return best_paths

            # get min vertex
            current_best = np.nanargmin(feasible_vertices.flatten())
            (x2, x3) = current_best // arr_len, current_best % arr_len
            x1 = min_shift_dists[x2, x3]

            # compute path and add to set
            vertices_path = self.graph._combined_paths(x1, [x2, x3])
            best_paths.append(vertices_path)
            for (x, y) in vertices_path:
                aux_arr[x, y] = 0

        self.graph.time_logs["ksp"] = round(time.time() - tic, 3)
        logger.debug(f"compute KSP time: {self.graph.time_logs['ksp']}")
        return best_paths

    def min_set_intersection(self, k, thresh=0.5):
        """
        Greedy Find KSP algorithm

        Arguments:
            self.graph.start_inds, self.graph.dest_inds: vertices -->
            list with two entries
            k: int: number of paths to output
            max_intersection: ratio of vertices that are allowed to be
            contained in the previously computed SPs
        """
        assert 0 <= thresh <= 1, "threshold for min_set_intersection\
             must be between 0 and 1"

        tic = time.time()

        best_paths = [np.array(self.graph.best_path)]

        # sum both dists_ab and dists_ba, subtract inst because counted twice
        (min_node_dists, min_shift_dists) = self.compute_min_node_dists()

        # argsort
        stack_sorted = np.dstack(
            np.unravel_index(
                np.argsort(min_node_dists.ravel()), min_node_dists.shape
            )
        )[0]
        # iterate over edges from least to most costly
        sorted_dist_prev = 0
        for x, y in stack_sorted:
            sorted_dist = min_node_dists[x, y]
            if sorted_dist == sorted_dist_prev:
                # on the same path as the vertex before
                continue
            if sorted_dist == np.inf:
                break
            sorted_dist_prev = sorted_dist
            s = min_shift_dists[x, y]

            # get shortest path through this node
            # if self.graph.dists_ba[x1, x2, x3] == 0:
            # = 0 for inc edges of self.graph.dest_inds_inds (init of dists_ba)
            # continue
            vertices_path = self.graph._combined_paths(s, [x, y])
            vertices_path = np.array(vertices_path)
            # compute intersection with previous paths
            intersection_low_enough = ut_ksp.intersecting_ratio(
                best_paths, vertices_path, thresh
            )
            # if similarity < threshold, add
            if intersection_low_enough:
                best_paths.append(vertices_path)
                if len(best_paths) >= k:
                    break
        self.graph.time_logs["ksp"] = round(time.time() - tic, 3)
        logger(f"FIND KSP time: {self.graph.time_logs['ksp']}")
        return best_paths



import unittest
import numpy as np
from lion.utils.plotting import plot_paths
from lion.algorithms import (
    optimal_point_spotting, optimal_route, ksp_points, ksp_routes
)


class TestAPI(unittest.TestCase):

    test_instance = np.random.rand(100, 100)
    num_nans = 50
    forb_x = (np.random.rand(num_nans) * 100).astype(int)
    forb_y = (np.random.rand(num_nans) * 100).astype(int)
    test_instance[forb_x, forb_y] = np.nan

    # create configuration
    cfg = dict()
    cfg["start_inds"] = [6, 6]
    cfg["dest_inds"] = [94, 90]
    test_instance[tuple(cfg["start_inds"])] = np.random.rand(1)[0]
    test_instance[tuple(cfg["dest_inds"])] = np.random.rand(1)[0]

    def test_optimal_route(self) -> None:
        path = optimal_route(
            self.test_instance,
            self.cfg.copy(),
        )
        self.assertTrue(len(path) > 0)
        self.assertListEqual(list(path[0]), list(self.cfg["start_inds"]))
        self.assertListEqual(list(path[-1]), list(self.cfg["dest_inds"]))

        plot_paths(
            self.test_instance, [path],
            buffer=0,
            out_path="test_optimal_route.png"
        )

    def test_optimal_point_spotting(self) -> None:
        path = optimal_point_spotting(
            self.test_instance,
            self.cfg.copy(),
        )
        self.assertTrue(len(path) > 0)
        self.assertListEqual(list(path[0]), list(self.cfg["start_inds"]))
        self.assertListEqual(list(path[-1]), list(self.cfg["dest_inds"]))

        plot_paths(
            self.test_instance, [path],
            buffer=0,
            out_path="test_optimal_point_spotting.png"
        )
        cfg = self.cfg.copy()
        cfg["pipeline"] = [3, 1]
        path_pipeline = optimal_point_spotting(
            self.test_instance,
            cfg,
        )
        self.assertListEqual(path, path_pipeline)

    def test_ksp_routes(self) -> None:
        paths = ksp_routes(self.test_instance, self.cfg.copy(), 5)
        self.assertTrue(len(paths) == 5)
        plot_paths(self.test_instance, paths, out_path="test_route_ksp.png")

    def test_ksp_points(self) -> None:
        paths = ksp_points(self.test_instance, self.cfg.copy(), 5)
        self.assertTrue(len(paths) == 5)
        plot_paths(self.test_instance, paths, out_path="test_point_ksp.png")


if __name__ == '__main__':
    unittest.main()



import lion.utils.general as ut
import lion.utils.costs as ut_cost
from lion.utils.shortest_path import get_update_algorithm
from lion.fast_shortest_path import (
    sp_dag, sp_dag_reversed, topological_sort_jit, edge_costs
)
import logging
import numpy as np
import time
import pickle
from numba.typed import List

logger = logging.getLogger(__name__)


class AngleGraph():

    def __init__(
        self, instance, instance_corr, edge_instance=None, directed=True
    ):

        # initialiye edge instance
        self.instance = instance.copy()
        assert np.all(
            self.instance < np.inf
        ), "No infs allowed in instance input to AngleGraph"
        # construct instance with infs where hard constraints are:
        self.instance[np.where(instance_corr == 0)] = np.inf

        if edge_instance is None:
            self.edge_inst = instance
        else:
            self.edge_inst = edge_instance
        self.x_len, self.y_len = instance_corr.shape
        self.time_logs = {}
        self.directed = directed

    def set_shift(
        self,
        start,
        dest,
        point_dist_min=3,
        point_dist_max=5,
        max_angle=np.pi / 2,
        **kwargs
    ):
        """
        Initialize shift variable by getting the donut values

        Arguments:
            start, dest: list containing X and Y coordinate of source / dest
            point_dist_min, point_dist_max: min and max distance of points
            max_angle: Maximum angle of edges to vec
        """
        self.start_inds = np.asarray(start)
        self.dest_inds = np.asarray(dest)
        vec = self.dest_inds - self.start_inds
        shifts = ut.get_half_donut(
            point_dist_min, point_dist_max, vec, angle_max=max_angle
        )
        shift_angles = [ut.angle_360(s, vec) for s in shifts]
        # sort the shifts
        self.shifts = np.asarray(shifts)[np.argsort(shift_angles)]
        self.shift_tuples = self.shifts

        # determine whether the graph is directed acyclic
        self.is_dag = max_angle <= np.pi / 2

        # construct bresenham lines
        shift_lines = List()
        for shift in self.shifts:
            line = ut.bresenham_line(0, 0, shift[0], shift[1])
            shift_lines.append(np.array(line))
        self.shift_lines = shift_lines

        # construct shift values for diagonals
        self.shift_costs = np.array([np.linalg.norm(s) for s in self.shifts])

    def _sort_cells_topologically(self):
        """
        Interface to topological_sort_jit - sort the vertices in the graph
        topologically to iterate over them in the correct order
        """
        tic = time.time()
        # SORT --> Make stack
        visit_points = (self.instance < np.inf).astype(int)
        initial_pos2node = np.zeros(visit_points.shape) - 1
        # run topologial sort --> fill pos2node
        initial_pos2node, _ = topological_sort_jit(
            self.dest_inds[0], self.dest_inds[1],
            np.asarray(self.shifts) * (-1), visit_points, initial_pos2node, 0
        )
        logger.info(f"topological sort in : {round(time.time() - tic, 3)}")
        return initial_pos2node

    def add_nodes(self):
        """
        Initialize distances and predecessors, sort vertices topologically
        """
        # sort vertices topologically
        initial_pos2node = self._sort_cells_topologically()

        # build stack from pos2node (sort it)
        self.stack_array = np.dstack(
            np.unravel_index(
                np.argsort(initial_pos2node.ravel()), initial_pos2node.shape
            )
        )[0]
        start_point = np.where(
            np.all(self.stack_array == self.start_inds, axis=1)
        )[0][0]
        self.stack_array = (self.stack_array[start_point:]).astype(int)
        logger.debug(f"number of vertices in stack: {len(self.stack_array)}")

        # build pos2node
        self.pos2node = (
            initial_pos2node - start_point +
            len(initial_pos2node[initial_pos2node == -1])
        ).astype(int)
        self.pos2node[self.pos2node < 0] = -1

        # check whether start can be reached from dest at all
        # if yes, overwrite stack array etc, because it's wrong
        if initial_pos2node[tuple(self.start_inds)] == -1:
            logger.warning("start not reachable from target, empty path")
            self.stack_array = np.array([self.start_inds, self.dest_inds])
            self.pos2node = np.zeros(initial_pos2node.shape).astype(int) - 1
            self.pos2node[tuple(self.start_inds)] = 0
            self.pos2node[tuple(self.dest_inds)] = 1

        # initializes dists and predecessors
        self.dists = np.zeros(
            (len(self.stack_array), len(self.shifts))
        ) + np.inf
        self.dists[0, :] = 0
        self.n_pixels = self.x_len * self.y_len
        self.n_nodes = len(self.stack_array)
        self.n_edges = len(self.shifts) * len(self.dists)
        logger.info(
            f"Number of edges: {len(self.stack_array)} x {len(self.shifts)}"
        )

    def set_edge_costs(
        self,
        angle_weight=0,
        max_angle_lg=np.pi,
        angle_cost_function='linear',
        between_points_allowed=True,
        **kwargs
    ):
        """
        Define the resistances / costs to represent on graph edges
        (Combines class-wise resistances and angle weights)

        Arguments:
            layer_classes: List of strings, names of cost categories
            class_weights: List of same length as layer_classes, corresponding
                weights (normalized automatically, can be any positive number)
            angle_weight: Importance of angle costs compared to resistances
                (=0 means only resistance is optimized, =1 means only angles
                are minimized, i.e. output will be straightest line possible)
            max_angle_lg: maximum angle between a adjacent edges on the path
            angle_cost_function: Currently implemented "linear" and "discrete"
                        Function defines the cost per angle, implemented in
                        utils/general.py (function compute_angle_cost)
        """
        assert 0 <= angle_weight <= 1, "angle weight must be between 0 and 1"

        # set weights and add angle weight
        self.resistance_weight = (1 - angle_weight)
        # set angle weight
        self.angle_weight = angle_weight

        # in precomute angles, it is multiplied with angle weights
        self.angle_cost_array = self._precompute_angles(
            max_angle_lg, angle_cost_function
        )
        # multiply with angle weights, need to prevent that not inf * 0
        non_inf = self.angle_cost_array < np.inf
        self.angle_cost_array[
            non_inf] = self.angle_cost_array[non_inf] * self.angle_weight

        # multiply resistances with other corresponding weight
        non_inf = self.instance < np.inf
        self.instance[non_inf
                      ] = self.instance[non_inf] * self.resistance_weight

        # If it is not allowed to traverse forbidden areas (i.e. forbidden area)
        # is between two points), then transform edge instance accordingly
        if not between_points_allowed:
            self.edge_inst[self.instance == np.inf] = np.inf

    def _precompute_angles(self, max_angle_lg, angle_cost_function):
        """
        Helper function to precompute the angle costs for all tuples of edges
        Arguments:
            max_angle_lg: maximum feasible angle
            angle_cost_function: funct to compute cost dependent on the angle
                        currently implemented: linear and one discrete option
        """
        self.angle_cost_function = angle_cost_function
        tic = time.time()

        # normalize (otherwise need to normalize in quadratic loop)
        shift_arrs = [np.asarray(s) for s in self.shifts]
        norm_shifts = [s / np.linalg.norm(s) for s in shift_arrs]
        # compute raw angle values
        angles_raw = np.array(
            [
                [ut.angle(s2, s1, normalize=False) for s1 in norm_shifts]
                for s2 in norm_shifts
            ]
        )
        # compute feasible maximum value
        max_angle = np.max(angles_raw[angles_raw <= max_angle_lg])
        self.angle_norm_factor = max_angle

        # compute angle costs (and normalize)
        slen = len(self.shifts)
        angles_all = np.array(
            [
                [
                    ut.compute_angle_cost(
                        angles_raw[i, j],
                        self.angle_norm_factor,
                        mode=self.angle_cost_function
                    ) for i in range(slen)
                ] for j in range(slen)
            ]
        )
        # greater than 1 (normalized) means infeasible angle
        angles_all[angles_all > 1] = np.inf

        self.time_logs["compute_angles"] = round(time.time() - tic, 3)
        return angles_all

    # --------------------------------------------------------------------
    # SHORTEST PATH COMPUTATION

    def build_source_sp_tree(self, edge_weight=0, **kwargs):
        self.edge_weight = edge_weight
        shift_norms = np.array([np.linalg.norm(s) for s in self.shifts])
        if np.any(shift_norms == 1):
            logger.warning("raster approach - edge weight set to zero")
            self.edge_weight = 0

        shift_norms = [np.linalg.norm(s) for s in self.shifts]
        tic = time.time()
        # precompute edge costs
        self.edge_cost = np.zeros(self.dists.shape) + np.inf
        self.edge_cost = edge_costs(
            self.stack_array, self.pos2node, np.array(self.shifts),
            self.edge_cost, self.instance, self.edge_inst, self.shift_lines,
            self.shift_costs, self.edge_weight
        )
        logger.debug(f"Computed edge costs in {time.time() - tic}")
        tic = time.time()
        # prepare for discrete if it is a discrete angle cost function:
        self.algorithm, self.args = get_update_algorithm(
            self.angle_cost_function, self.angle_cost_array, self.shifts
        )

        # RUN - either directed acyclic or BF algorithm
        if self.is_dag:
            self.dists, self.preds = sp_dag(
                self.stack_array, self.pos2node, np.array(self.shifts),
                self.angle_cost_array, self.dists, self.edge_cost,
                self.algorithm, self.args
            )
        else:
            raise NotImplementedError(
                "Angle shortest path not implemented for cyclic graph.\
                    Please set max_angle <= np.pi / 2 in config"
            )

        self.time_logs["shortest_path"] = round(time.time() - tic, 3)
        logger.debug(f"time single SP: {round(time.time() - tic, 3)}")

    # ----------------------------------------------------------------------
    # REVERSED TREE FOR KSP

    def build_dest_sp_tree(self):
        """
        Compute costs from dest to all edges
        """
        tic = time.time()

        # initialize dists array
        self.dists_ba = np.zeros(self.dists.shape) + np.inf
        # this time need to set all incoming edges of dest to zero
        # d0, d1 = self.dest_inds
        # for s, (i, j) in enumerate(self.shifts):
        #     pos_index = self.pos2node[d0 + i, d1 + j]
        #     self.dists_ba[pos_index, s] = 0

        # compute distances: new method because out edges instead of in
        self.dists_ba, self.preds_ba = sp_dag_reversed(
            self.stack_array, self.pos2node,
            np.array(self.shifts) * (-1), self.angle_cost_array, self.dists_ba,
            self.edge_cost, self.algorithm, self.args
        )
        self.time_logs["shortest_path_tree"] = round(time.time() - tic, 3)
        logger.debug(f"done shortest_path_tree:{round(time.time() - tic, 3)}")
        # from lion.utils.plotting import angle_graph_display_dists
        # self.angle_graph_display_dists(self.dists_ba)
        # distance in ba: take IN edges to source, by computing in neighbors
        # take their first dim value (out edge to source) + source val
        (s0, s1) = self.start_inds
        neigh_inds = [self.pos2node[s0 + i, s1 + j] for (i, j) in self.shifts]
        start_dests = [
            self.dists_ba[neigh_inds[s], s] if neigh_inds[s] >= 0 else np.inf
            for s, (i, j) in enumerate(self.shifts)
        ]
        d_ba = np.min(start_dests)

        d_ab = np.min(self.dists[self.pos2node[tuple(self.dest_inds)], :])
        assert np.isclose(
            d_ba, d_ab
        ), "start to dest != dest to start " + str(d_ab) + " " + str(d_ba)

    def _combined_paths(self, best_shift, best_edge):
        """
        Compute path through one specific edge (with bi-directed predecessors)

        Arguments:
            best_edge: the vertex of the edge
            best_shift: the neighbor index of the edge
        """
        # compute path from start to middle point - incoming edge
        best_edge = np.array(best_edge)
        path_ac = self.get_sp_start_shift(best_edge, best_shift)
        # compute path from middle point to dest - outgoing edge
        path_cb = self.get_sp_dest_shift(best_edge, best_shift)
        # concatenate
        together = np.concatenate(
            (np.flip(np.array(path_ac), axis=0), np.array(path_cb)[1:]),
            axis=0
        )
        return together

    # ---------------------------------------------------------------------
    # Functions to output path (backtrack) and corresponding costs

    def transform_path(self, path):
        # raw_resist = np.array([[self.instance[p[0], p[1]]] for p in path])

        # compute angle costs
        ang_costs = ut_cost.compute_angle_costs(
            path, self.angle_norm_factor, mode=self.angle_cost_function
        )

        # compute the geometric path costs
        path_costs = ut_cost.compute_geometric_costs(
            path, self.instance, self.edge_weight
        )
        # combine costs
        cost_sum = np.sum(path_costs) + self.angle_weight * np.sum(ang_costs)
        return np.asarray(path).tolist(), np.array(path_costs), cost_sum

    def get_shortest_path(self):
        dest_ind_stack = self.pos2node[tuple(self.dest_inds)]
        if not np.any(self.dists[dest_ind_stack, :] < np.inf):
            logger.warning("WARNING: Empty path!")
            return [], [], 0
        tic = time.time()

        # backtrack from dest to start from predecessor maps
        min_shift = np.argmin(self.dists[dest_ind_stack, :])
        path = self.get_sp_start_shift(self.dest_inds, min_shift)

        path = np.flip(np.asarray(path), axis=0)
        self.best_path = path
        self.time_logs["path"] = round(time.time() - tic, 3)
        return path.tolist()

    def get_sp_start_shift(self, vertex_coords, min_shift):
        """
        Reconstruct paths from distance and predecessor map with REVERSED
        edge directions
        @param
            vertex_coords: coordinates of current point
            min_shift: optimal predecessor shift at target (to trace back)
        @returns:
            list of X, Y, Z coordinates
        """
        vertex_stack = self.pos2node[tuple(vertex_coords)]
        if not np.any(self.dists[vertex_stack, :] < np.inf):
            raise RuntimeWarning("empty path")
        curr_point = np.asarray(vertex_coords)
        my_path = [vertex_coords]
        while np.any(curr_point - self.start_inds):
            new_point = curr_point - self.shifts[int(min_shift)]
            # get next shift
            pred_ind = self.pos2node[tuple(curr_point)]
            min_shift = self.preds[pred_ind, int(min_shift)]
            my_path.append(new_point)
            curr_point = new_point
        return my_path

    def get_sp_dest_shift(self, vertex_coords, min_shift, dest_edge=False):
        """
        Reconstruct paths from distance and predecessor map with REVERSED
        edge directions
        @param
            vertex_coords: coordinates of current point
            min_shift: optimal predecessor shift at target (to trace back)
            dest_edge: Whether to use min_shift or not
        @returns:
            list of X, Y coordinates
        """
        if not dest_edge:
            vertex_coords_ind = self.pos2node[tuple(vertex_coords)]
            min_shift = self.preds_ba[vertex_coords_ind, int(min_shift)]
        curr_point = np.asarray(vertex_coords)
        my_path = [vertex_coords]
        while np.any(curr_point - self.dest_inds):
            # move on to next point
            new_point = curr_point + self.shifts[int(min_shift)]
            # get next shift
            pred_ind = self.pos2node[tuple(new_point)]
            min_shift = self.preds_ba[pred_ind, int(min_shift)]
            my_path.append(new_point)
            curr_point = new_point
        return my_path

    # ----------------------------------------------------------------------
    # Other auxiliary functions
    def save_graph(self, out_path):
        with open(out_path + ".dat", "wb") as outfile:
            pickle.dump((self.dists, self.preds), outfile)

    def _helper_list(self):
        tmp_list = List()
        tmp_list_inner = List()
        tmp_list_inner.append(0)
        tmp_list_inner.append(0)
        tmp_list.append(tmp_list_inner)
        return tmp_list

    # -----------------------------------------------------------------------
    # INTERFACE

    def _check_start_dest(self, **kwargs):
        # assert that start and dest are provided and in project bounds
        try:
            self.start_inds = kwargs["start_inds"]
            self.dest_inds = kwargs["dest_inds"]
        except KeyError:
            raise RuntimeError(
                "Must specify start_inds and dest_inds in cfg dict!"
            )
        assert len(self.start_inds) == 2, "start inds must be of length 2"
        assert len(self.dest_inds) == 2, "dest inds must be of length 2"
        assert all(
            [
                isinstance(self.start_inds[i], int)
                or float(self.start_inds[i]).is_integer() for i in range(2)
            ]
        ), "start inds must be integer!"
        assert all(
            [
                isinstance(self.dest_inds[i], int)
                or float(self.dest_inds[i]).is_integer() for i in range(2)
            ]
        ), "dest inds must be integer!"
        assert self.instance[tuple(
            self.start_inds
        )] < np.inf, "Problem: Start coordinates are not in project region"
        assert self.instance[tuple(
            self.dest_inds
        )] < np.inf, "Problem: Target coordinates are not in project region"
        self.start_inds = np.asarray(self.start_inds).astype(int)
        self.dest_inds = np.asarray(self.dest_inds).astype(int)

        instance_shape = np.asarray(self.instance.shape)
        assert np.all(np.asarray(self.start_inds) < instance_shape) and np.all(
            np.asarray(self.dest_inds) < instance_shape
        ), "start or dest not in project region!"

    def single_sp(self, **kwargs):
        """
        Function for full processing to yield shortest path
        Necessary parameters:
            start_inds: list of two cell coordinates
            dest_inds: list of two cell coordinates
        Optional parameters:
            point_dist_min: min cell distance of neighboring points (default 3)
            point_dist_max: min cell distance of neighboring points (default 5)
            angle_weight: Importance of angle costs compared to resistances
                (=0 means only resistance is optimized, =1 means only angles
                are minimized, i.e. output will be straightest line possible)
            edge_weight: importance of costs of the cells between points vs 
                cost at the point themselves (default 0 --> only points matter)
            max_angle: maximum deviation in angle from the straight connection
                       from start to end (default: pi/2)
            max_angle_lg: maximum angle at a point (default: pi/2)
        """
        # assert that start and dest exist in kwargs and are in project region
        self._check_start_dest(**kwargs)

        # initialize donut ring and edge costs
        self.set_shift(self.start_inds, self.dest_inds, **kwargs)
        logger.debug("1) Initialize shifts and instance (corridor)")
        self.set_edge_costs(**kwargs)
        # add vertices
        self.add_nodes()
        logger.debug("2) Initialize distances to inf and predecessors")
        # MAIN ALGORITHM
        self.build_source_sp_tree(**kwargs)
        logger.debug("3) Compute source shortest path tree")
        logger.debug(
            f"number of vertices: {self.n_nodes} and edges: {self.n_edges}"
        )

        # get actual best path
        path = self.get_shortest_path()
        logger.debug("4) shortest path computed")
        return path

    def sp_trees(self, **kwargs):
        """
        Compute shortest path trees from both directions (Eppstein distances)
        necessary for finding multiple paths
        """
        # Build shortest path tree rooted in source
        path = self.single_sp(**kwargs)
        # Build shortest path tree rooted in target
        self.build_dest_sp_tree()
        return path



"""
Methods to compute the shortest path or multiple shortest paths
given an instance and a configuration file

Main parameters:

instance: 2D numpy array of resistances values (float or int)
          If cells are forbidden, set them to NaN or a value x and specify x
          as cfg["forbidden_val"] = x.

cfg - configuration: Dict with the following neceassay and optional parameters
    + start_inds: list of two cell coordinates
    + dest_inds: list of two cell coordinates

    - forbidden_val: value indicating that a cell is forbidden / outside of
            the project region (can be int, np.nan, np.inf ... )
    - point_dist_min: minimum cell distance of neighboring points (default 3)
    - point_dist_max: minimum cell distance of neighboring points (default 5)
    - angle_weight: how important is the angle (default 0)
    - edge_weight: importantance of costs between points compared to points
            themselves (default 0 --> only the cost at the points matters)
    - max_angle: maximum deviation in angle from the straight connection from
            start to end (default: pi/2)
    - max_angle_lg: maximum angle at a point (default: pi)
    - angle_cost_function: 'linear' and 'discrete' are implemented
    - memory_limit: Maximum number of edges that is allowed (default: 50 Mio)
            If the number of edges is higher, an iterative procedure is used.
            Relation of edges to actual memory:
            For single SP, space for two float arrays with #edges cells are
                allocated -> 2 * #edges * 8 bytes (float64) = 16 x #edges bytes
            For multi SP there will be four float arrays with #edges cells
                -> 32 x #edges bytes
    - pipeline: List of decreasing positive integers, ending with 1
            The pipeline in an iterative approach defines the downsampling
            factors for each step. By default, it is set automatically based on
            the memory limit. It can however be set manually as well, e.g.
            [3,1] means downsample by factor of 3, compute optimal
            path, reduce region of interest to a corridor around
            optimal path (corridor width is computed automatically based on the
            memory_limit) then downsample by factor of 1 (aka full resolution).
            There is no support for setting the corridor width manually because
            it does not make sense to make it smaller than it could be
    - between_points_allowed: If True, then forbidden areas can still be
            traversed, i.e. two points can be placed around a forbidden area
            If False, then forbidden areas can not be traversed either
    - diversity_threshold:
        FOR KSP.ksp:
            Minimum diversity of next path from previous paths in cell
            distance. E.g. if thresh = 200, each path will be at least 200
            cells away at one point from each other path.
            If None, it is set by default to 1/20 of the instance size

        FOR KSP.min_set_intersection:
            maximum intersection of the new path with all previous points
            Must be between 0 and 1. E.g. if 0.2, then at most 20% of cells
            are shared between one path and the other paths
"""

import numpy as np
from lion.angle_graph import AngleGraph
from lion.ksp import KSP
import lion.utils.costs as ut_cost
import lion.utils.general as ut_general
import time
import logging

logger = logging.getLogger(__name__)


def _initialize_graph(instance, cfg):
    """
    Auxiliary method to preprocess the instance and initialize the graph
    @params:
        instance: 2D numpy array of resistances values (float or int)
                  If cells are forbidden, set them to NaN or a value x and
                  specify x in cfg["forbidden_val"].
        cfg: configuration as specified on top of this file
    @returns:
        graph: AngleGraph object initialized with instance
        cfg: updated configuration file
    """
    forbidden_val = cfg.get("forbidden_val", np.nan)
    logger.info(f"forbidden val: {forbidden_val}")

    # make forbidden region array
    project_region = np.ones(instance.shape)
    project_region[np.isnan(instance)] = 0
    project_region[instance == forbidden_val] = 0

    # normalize instance -- necessary to have comparable angle weight
    normal_vals = instance[
        np.logical_and(instance != forbidden_val, ~np.isnan(instance))]
    assert np.all(
        normal_vals < np.inf
    ), "check forbidden_val parameter in cfg, it is\
         not inf but there are inf values in array"

    # fill values in instance
    instance[project_region == 0] = np.max(normal_vals)
    non_zero_add = 1e-8  # instance must not contain zeros
    # normalize to values between non_zero_add and 1
    instance = non_zero_add + (1 - non_zero_add) * (
        instance - np.min(normal_vals)
    ) / (np.max(normal_vals) - np.min(normal_vals))
    assert np.min(instance) > 0 and np.isclose(
        np.max(instance), 1
    ), "Minimum must be greater than zero and maximum ~1 after normalizing"

    # initialize graph
    graph = AngleGraph(instance, project_region)

    return graph, cfg


def optimal_route(instance, cfg):
    """
    Compute the (angle-) optimal path through a grid
    @params:
        instance: 2D np array of resistances (see details top of file)
        cfg: config dict, must contain start and dest (see details top of file)
        see top of file for more details
    @returns:
        a single optimal path (list of X Y coordinates)
        or empty list if no path exists
    """
    # initialize graph
    graph, cfg = _initialize_graph(instance, cfg)

    # set the ring to the 8-neighborhood
    cfg["point_dist_min"] = 0.9
    cfg["point_dist_max"] = 1.5

    # compute path
    tic_raster = time.time()
    path = graph.single_sp(**cfg)

    logger.info(f"Overall timefor optimal route: {time.time() - tic_raster}")
    logger.info(f"time logs: {graph.time_logs}")

    return path


def optimal_point_spotting(
    instance, cfg, corridor=None, k=1, algorithm=KSP.ksp
):
    """
    Compute the (angle-) optimal point spotting
    @params:
        instance: 2D np array of resistances (see details top of file)
        cfg: config dict, must contain start and dest (see details top of file)
        corridor: relevant region --> either
            - a path, e.g. output of optimal_route, or
            - a 2D array with 0: do not consider, 1: consider
    @returns:
        a single optimal path of points (list of X Y coordinates)
        or empty list if no path exists
    """
    # initialize graph
    graph, cfg = _initialize_graph(instance, cfg)
    # initial project region
    original_inst = graph.edge_inst.copy()
    original_corr = (graph.instance < np.inf).astype(int)
    # initialize corridor
    if corridor is None:
        corridor = np.ones(original_corr.shape)

    # estimate how many edges the graph will have:
    graph.set_shift(cfg["start_inds"], cfg["dest_inds"], **cfg)
    orig_shifts = len(graph.shift_tuples)
    instance_vertices = np.sum(original_corr * corridor > 0)

    mem_limit = cfg.get("memory_limit", 5e7)
    # define pipeline
    pipeline = cfg.get(
        "pipeline",
        ut_general.get_pipeline(instance_vertices, orig_shifts, mem_limit)
    )
    assert all(
        pipeline[i] > pipeline[i + 1] for i in range(len(pipeline) - 1)
    ), "pipeline must consist of decreasing downsampling factors"
    assert pipeline[
        -1] == 1, "last factor in pipeline must be 1 (= no downsampling)"

    # execute pipeline
    logger.info(f"Pipeline set to: {pipeline}")

    # execute iterative shortest path computation
    for pipe_step, factor in enumerate(pipeline):
        assert isinstance(factor, int) or float(factor).is_integer(
        ), "downsampling factors in pipeline must be integers"
        logger.info(
            f"---------- Start {pipe_step+1}th step {factor} ---------------"
        )
        # rescale and set parameters accordingly
        corridor = (corridor * original_corr > 0).astype(int)
        current_instance, current_corridor, current_cfg = ut_general.rescale(
            original_inst, corridor, cfg, factor
        )
        # run shortest path computation
        graph = AngleGraph(current_instance, current_corridor)
        if k > 1:
            paths = _run_ksp(graph, current_cfg, k, algorithm=algorithm)
            paths = [np.array(path) * factor for path in paths]
        else:
            path = graph.single_sp(**current_cfg)
            paths = [np.asarray(path) * factor]
        logger.debug(f"got {len(paths)} paths in this step")

        # compute next corridor
        if pipe_step < len(pipeline) - 1:
            corridor = ut_general.pipeline_corridor(
                paths, instance.shape, orig_shifts, mem_limit,
                pipeline[pipe_step + 1]
            )
    if k == 1:
        return path
    else:
        return paths


# ----------------------------------- KSP  ---------------------------------
def _run_ksp(graph, cfg, k, algorithm=KSP.ksp):
    """
    Build the shortest path trees and compute k diverse shortest paths
    Arguments:
        graph: AngleGraph object that was initialized with the instance
               (as passed from ksp_routes or ksp_points)
        cfg: See beginning of file for possible parameters
        k: Number of diverse alternatives to compute. Attention: Might output
            less than k paths if no further sufficiently diverse path is found
        algorithm: currently implemented in lion/ksp.py :
                KSP.ksp: Interatively find next shortest path that is at
                         least <thresh> cells from the previously found paths
                KSP.min_set_intersection: Iteratively find next shortest path
                         that shares at most <thresh> cells with the previous
    """

    def set_thresh_automatically():
        # helper method to set the diversity threshold dependent on inst size
        if algorithm == KSP.min_set_intersection:
            thresh = 0.3
        else:
            inst_size = min([graph.instance.shape[0], graph.instance.shape[1]])
            thresh = int(inst_size / 10)
        logger.info(f"set diversity treshold automatically to: {thresh}")
        return thresh

    thresh = cfg.get("diversity_threshold", set_thresh_automatically())
    logger.debug(f"diversity threshold is {thresh}")

    # construct sp trees
    tic = time.time()
    _ = graph.sp_trees(**cfg)
    # compute k shortest paths
    ksp_processor = KSP(graph)
    ksp_paths = algorithm(ksp_processor, k, thresh=thresh)
    logger.info(f"Time for run ksp: {time.time() - tic}")
    return ksp_paths


def ksp_routes(instance, cfg, k, algorithm=KSP.ksp):
    """
    Compute the (angle-) optimal k diverse shortest paths through a grid
    @params:
        instance: 2D np array of resistances (see details top of file)
        cfg: config dict, must contain start and dest (see details top of file)
        k: number of paths to compute
        algorithm: see doc of run_ksp
    @returns:
        A list of paths (each path is again a list of X Y coordinates)
    """
    # initialize graph
    graph, cfg = _initialize_graph(instance, cfg)

    # set the ring to the 8-neighborhood
    cfg["point_dist_min"] = 1
    cfg["point_dist_max"] = 1.5

    # run algorithm
    return _run_ksp(graph, cfg, k, algorithm=algorithm)


def ksp_points(instance, cfg, k, algorithm=KSP.ksp):
    """
    Compute the (angle-) optimal k diverse shortest path of pointS
    @params:
        instance: 2D np array of resistances (see details top of file)
        cfg: config dict, must contain start and dest (see details top of file)
        k: number of paths to compute
        algorithm: see doc of run_ksp
    @returns:
        A list of paths (each path is again a list of X Y coordinates)
    """
    return optimal_point_spotting(instance, cfg, k=k, algorithm=algorithm)



import numpy as np
import logging

try:
    import matplotlib.pyplot as plt
    _PLOT_ENABLED = True
except ImportError:
    _PLOT_ENABLED = False

logger = logging.getLogger(__name__)


def angle_graph_display_dists(self, edge_array, func=np.min, name="dists"):
    if not _PLOT_ENABLED:
        logger.warning('No plotting library available')
        return

    arr = np.zeros(self.pos2node.shape)
    for i in range(len(self.pos2node)):
        for j in range(len(self.pos2node[0])):
            ind = self.pos2node[i, j]
            if ind >= 0:
                arr[i, j] = func(edge_array[ind, :])
    plt.imshow(arr)
    plt.colorbar()
    plt.savefig(name + ".png")


def plot_paths(instance, paths, buffer=0, out_path="test_path.png"):
    if not _PLOT_ENABLED:
        logger.warning('No plotting library available')
        return
    expanded = np.expand_dims(instance, axis=2)
    expanded = np.tile(expanded, (1, 1, 3))  # overwrite instance by tiled one
    # colour nodes in path in red
    for path in paths:
        for (x, y) in path:
            expanded[x - buffer:x + buffer + 1, y - buffer:y + buffer +
                     1] = [0.9, 0.2, 0.2]  # colour red
    # plot and save
    plt.figure(figsize=(25, 15))
    plt.imshow(np.swapaxes(expanded, 1, 0))
    plt.savefig(out_path, bbox_inches='tight')


def plot_dists(self, dists, pos2node, save_name):
    arr = np.zeros(pos2node.shape)
    for i in range(len(pos2node)):
        for j in range(len(pos2node)):
            if pos2node[i, j] >= 0:
                arr[i, j] = np.min(dists[pos2node[i, j]])
    plt.imshow(arr)
    plt.savefig(save_name)



import unittest
import numpy as np
from lion.angle_graph import AngleGraph
from types import SimpleNamespace
from lion.fast_shortest_path import sp_dag
from lion.utils.shortest_path import (
    update_default, update_linear, update_discrete
)


class TestUpdateAlgs(unittest.TestCase):

    expl_shape = (100, 100)
    instance = np.random.rand(*expl_shape)
    corridor = (np.random.rand(*expl_shape) > 0.05).astype(int)

    # create configuration
    cfg = dict()
    start_inds = np.array([6, 6])
    dest_inds = np.array([94, 96])
    corridor[tuple(start_inds)] = 1
    corridor[tuple(dest_inds)] = 1
    cfg["start_inds"] = start_inds
    cfg["dest_inds"] = dest_inds
    cfg["point_dist_min"] = 10
    cfg["point_dist_max"] = 15
    cfg["layer_classes"] = ["dummy_class"]
    cfg["class_weights"] = [1]

    def test_linear(self) -> None:
        """ LINEAR """
        self.cfg["angle_cost_function"] = "linear"
        self.cfg["max_angle_lg"] = np.pi
        graph = AngleGraph(self.instance, self.corridor)
        _ = graph.single_sp(**self.cfg)
        gt_angle_cost_arr = graph.angle_cost_array.copy()
        gt_dists = graph.dists.copy()
        graph.dists = np.zeros(
            (len(graph.stack_array), len(graph.shifts))
        ) + np.inf
        graph.dists[0, :] = 0
        graph.preds = np.zeros(graph.dists.shape) - 1
        graph.angle_cost_function = "some_non_existant"
        graph.build_source_sp_tree(**self.cfg)

        self.assertTrue(
            np.all(np.isclose(gt_angle_cost_arr, graph.angle_cost_array))
        )
        self.assertTrue(np.all(np.isclose(gt_dists, graph.dists)))

        # OTHER WAY: FROM NOTEBOOK
        # stack_saved = graph.stack_array.copy()
        # self.assertTrue(np.all(graph.angle_cost_array < np.inf))
        # dists_new = np.zeros(
        #     (len(graph.stack_array), len(graph.shifts))
        # ) + np.inf
        # dists_new[0, :] = 0
        # preds_new = np.zeros(dists_new.shape) - 1

        # gt_dists, _ = sp_dag(
        #     stack_saved, graph.pos2node,
        #     np.array(graph.shifts), graph.angle_cost_array, dists_new.copy(),
        #     preds_new.copy(), graph.edge_cost, update_default,
        #     (graph.angle_cost_array)
        # )

        # linear_dists, _ = sp_dag(
        #     stack_saved, graph.pos2node,
        #     np.array(graph.shifts), graph.angle_cost_array, dists_new.copy(),
        #     preds_new.copy(), graph.edge_cost, update_linear,
        #     (graph.angle_cost_array)
        # )

        # self.assertTrue(np.all(gt_dists == linear_dists))

    def test_discrete(self) -> None:
        """ DISCRETE """
        self.cfg["angle_cost_function"] = "discrete"
        self.cfg["max_angle_lg"] = np.pi / 4
        graph = AngleGraph(self.instance, self.corridor)
        _ = graph.single_sp(**self.cfg)
        gt_angle_cost_arr = graph.angle_cost_array.copy()
        gt_dists = graph.dists.copy()
        graph.dists = np.zeros(
            (len(graph.stack_array), len(graph.shifts))
        ) + np.inf
        graph.dists[0, :] = 0
        graph.preds = np.zeros(graph.dists.shape) - 1
        graph.angle_cost_function = "some_non_existant"
        graph.build_source_sp_tree(**self.cfg)

        self.assertTrue(np.all(gt_angle_cost_arr == graph.angle_cost_array))
        self.assertTrue(np.all(np.isclose(gt_dists, graph.dists)))


if __name__ == '__main__':
    unittest.main()



import numpy as np
from numba import jit
import logging

logger = logging.getLogger(__name__)


@jit(nopython=True)
def compute_eucl(path1, path2, mode="mean"):
    """
    Compute Euclideean distance between two paths
    @param mode: if "mean", then compute the average of the minimum distances
                of all vertices of path1 to path2 and the other way round,
                if "max" then the maximum of minimum distances
    @returns: float indicating the distance
    """
    min_dists_out = np.zeros(len(path1))
    for p1 in range(len(path1)):
        min_dists = np.zeros(len(path2))
        for p2 in range(len(path2)):
            min_dists[p2] = np.linalg.norm(path1[p1] - path2[p2])
        min_dists_out[p1] = np.min(min_dists)
    if mode == "mean":
        return np.mean(min_dists_out)
    elif mode == "max":
        return np.max(min_dists_out)
    else:
        raise NotImplementedError("only mean or max distance are possible")


@jit(nopython=True)
def fast_dilation(path_points, arr_shape, iters=50):
    """
    Compute distance transform - indicates dist of each cell to a set of paths
    @params:
        path_points : 2D array of shape (n, 2) of cells on previous paths
        arr_shape : Tuple of integers, shape of raster
        iters : Maximum distance until which is should be computed
    @returns:
        2D raster of shape arr_shape containing the INVERTED distances:
        0 means more than iters distance, iters means distance 0
    """
    arr = np.zeros(arr_shape)
    if iters > arr_shape[0] and iters > arr_shape[1]:
        # logger.warn("corridor width was larger than instance size")
        return np.ones(arr_shape)
    for i in range(len(path_points)):
        arr[path_points[i][0], path_points[i][1]] = 1

    start_x, end_x = (np.min(path_points[:, 0]), np.max(path_points[:, 0]))
    start_y, end_y = (np.min(path_points[:, 1]), np.max(path_points[:, 1]))
    # todo: must be in array bounds
    for i in range(iters):
        arr_prev = arr.copy()
        for x in range(start_x - 1, end_x + 2):
            for y in range(start_y - 1, end_y + 2):
                if arr_prev[x, y] > 0 or arr_prev[x - 1, y] > 0 or (
                    arr_prev[x + 1, y] > 0
                ) or arr_prev[x, y - 1] > 0 or arr_prev[x, y + 1] > 0:
                    arr[x, y] += 1
        if start_x > 1:
            start_x -= 1
        if end_x < arr_shape[0] - 3:
            end_x += 1
        if start_y > 1:
            start_y -= 1
        if end_y < arr_shape[1] - 3:
            end_y += 1

    return arr


def intersecting_ratio(path_list, current_path, max_intersection):
    """
    Compute whether the new path is intersecting more than max_intersection
    with the paths in path_list

    path_list: list of arrays of shape (n,2) containing X Y coords of paths
    current_path: 2Darray of shape (m,2)
    max_intersection: float between 0 and 1 indicating intersection threshold
    """
    convert_num = np.max(current_path)
    current_path_converted = (
        current_path[:, 0] * convert_num + current_path[:, 1]
    )

    for prev_path_ind in range(len(path_list)):
        # get array
        prev_path = path_list[prev_path_ind]
        prev_path_converted = prev_path[:, 0] * convert_num + prev_path[:, 1]
        _, inds_intersection, _ = np.intersect1d(
            current_path_converted,
            prev_path_converted,
            assume_unique=True,
            return_indices=True
        )
        # delete the ones that have already been found
        current_path_converted = np.delete(
            current_path_converted, inds_intersection
        )

        # If above threshold already, then return
        if 1 - len(current_path_converted
                   ) / len(current_path) > max_intersection:
            return False

    # For all paths the intersection has stayed sufficiently low
    return True


def get_sp_from_preds(pred_map, curr_vertex, start_vertex):
    """
    Compute path from start_vertex to curr_vertex from the predecessor map
    Arguments:
        pred_map: map / dictionary with predecessor for each vertex
        curr_vertex: integer denoting any vertex
        start_vertex: integer denoting start vertex
    returns:
        list of vertices (integers)
    """
    path = [int(curr_vertex)]
    counter = 0
    while curr_vertex != start_vertex:
        curr_vertex = pred_map[curr_vertex]
        path.append(curr_vertex)
        if counter > 1000:
            print(path)
            raise RuntimeWarning("while loop for sp not terminating")
        counter += 1
    return path


def path_distance(p1, p2, mode="jaccard"):
    """
    Compute the distance between two paths
    NOTE: all modes in this method are valid metrics
    Arguments:
        p1,p1: two paths (lists of coordinates!)
        mode: jaccard: jaccard distance (1-IoU)
            eucl_mean: from all min eucl distances, take mean
            eucl_max: from all min eucl distances, take max
    """
    if mode == "jaccard":
        s1 = set([tuple(p) for p in p1])
        s2 = set([tuple(p) for p in p2])
        # s1,s2 = (set(list(p1)),set(list(p2)))
        return 1 - len(s1.intersection(s2)) / len(s1.union(s2))
    elif mode.startswith("euc"):
        p1 = np.array(p1).astype("float")
        p2 = np.array(p2).astype("float")
        eucl_mode = mode.split("_")[1]
        return max(
            [
                compute_eucl(p1, p2, mode=eucl_mode),
                compute_eucl(p2, p1, mode=eucl_mode)
            ]
        )
    else:
        raise NotImplementedError(
            "mode " + mode + " wrong, not implemented yet"
        )


def similarity(s1, s2, mode="IoU"):
    """
    Implements similarity metrics from Liu et al paper
    Arguments:
        s1,s2: SETS of path points
    """
    path_inter = len(s1.intersection(s2))
    if mode == "IoU":
        return path_inter / len(s1.union(s2))
    elif mode == "sim2paper":
        return path_inter / (2 * len(s1)) + path_inter / (2 * len(s2))
    elif mode == "sim3paper":
        return np.sqrt(path_inter**2 / (len(s1) * len(s2)))
    elif mode == "max_norm_sim":
        return path_inter / (max([len(s1), len(s2)]))
    elif mode == "min_norm_sim":
        return path_inter / (min([len(s1), len(s2)]))
    else:
        raise NotImplementedError("mode wrong, not implemented yet")


def pairwise_dists(collected_coords, mode="jaccard"):
    nr_paths = len(collected_coords)
    dists = np.zeros((nr_paths, nr_paths))
    for i in range(nr_paths):
        for j in range(i, nr_paths):
            dists[i, j] = path_distance(
                collected_coords[i], collected_coords[j], mode=mode
            )
            dists[j, i] = dists[i, j]
    return dists


def _flat_ind_to_inds(flat_ind, arr_shape):
    """
    Transforms an index of a flattened 3D array to its original coords
    """
    _, len2, len3 = arr_shape
    x1 = flat_ind // (len2 * len3)
    x2 = (flat_ind % (len2 * len3)) // len3
    x3 = (flat_ind % (len2 * len3)) % len3
    return (x1, x2, x3)


def evaluate_sim(ksp, metric):
    """
    evaluate ksp diversity according to several metric
    """
    ksp_paths = [k[0] for k in ksp]
    divs = []
    # compute pariwise path distances
    for i in range(len(ksp_paths)):
        for j in range(i + 1, len(ksp_paths)):
            divs.append(path_distance(ksp_paths[i], ksp_paths[j], mode=metric))
    return np.mean(divs)


def evaluate_cost(ksp):
    """
    Evaluate ksp with respect to the overall and maximal costs
    """
    ksp_all_costs = [k[2] for k in ksp]
    return [np.sum(ksp_all_costs), np.max(ksp_all_costs)]



import numpy as np
from lion.utils.general import (bresenham_line, compute_angle_cost, angle)
try:
    # import only required for the watershed transform, so imported
    # only of available in order to reduce dependencies
    from skimage.segmentation import watershed
    from skimage import filters
except ModuleNotFoundError:
    pass


def compute_edge_costs(path, instance):
    """
    Compute the raw edge costs, i.e. the average value of the crossed cells
    Arguments:
        path: list of tuples or 2D numpy array with path coordinates
        instance: 2D np array with resistance values
    Returns:
        A list with the same length as the path, containing the edge cost vals
        (last value is zero because less edges then vertices)
    """
    path = np.array(path)
    e_costs = []
    for p in range(len(path) - 1):
        point_list = bresenham_line(*tuple(path[p]), *tuple(path[p + 1]))
        e_costs.append(
            np.mean([instance[i, j] for (i, j) in point_list[1:-1]])
        )
    # to make it the same size as other costs
    e_costs.append(0)
    return np.array(e_costs)


def compute_geometric_costs(path, instance, edge_weight=0):
    """
    Compute geometric costs along the path
    Arguments:
        path: List or array of path corrdinates
        instance: 2D array of resistances
        edge_costs: 1D array, previously computed costs between points along
            the path
    Returns:
        List of geometric edge costs along the path
    """
    path = np.array(path)
    # compute the between point costs (bresenham line between points)
    edge_costs = np.zeros(len(path))
    if edge_weight != 0:
        edge_costs = compute_edge_costs(path, instance) * edge_weight

    geometric_costs = []
    for p in range(len(path) - 1):
        # between point costs
        bresenham_edge_dist = edge_costs[p]
        # compute distance inbetween
        shift_costs = np.linalg.norm(path[p] - path[p + 1])
        # compute geometric edge costs
        geometric_costs.append(
            shift_costs * (
                0.5 *
                (instance[tuple(path[p])] + instance[tuple(path[p + 1])]) +
                bresenham_edge_dist
            )
        )
    geometric_costs.append(0)
    return geometric_costs


def compute_angle_costs(path, angle_norm_factor=np.pi / 2, mode="linear"):
    """
    Compute the normalized angle costs along the path
    Arguments:
        path: list of tuples or 2D numpy array with path coordinates
        instance: 2D np array with resistance values
    Returns:
        List of same length as path containing angle cost values
        (first and last entry 0 because angle at s and t is irrelevant)
    """
    path = np.asarray(path)
    ang_out = [0]
    for p in range(len(path) - 2):
        vec1 = path[p + 1] - path[p]
        vec2 = path[p + 2] - path[p + 1]
        ang_out.append(
            compute_angle_cost(
                angle(vec1, vec2), angle_norm_factor, mode=mode
            )
        )
    ang_out.append(0)

    return ang_out


def compute_raw_angles(path):
    """
    Compute the raw angles along the path (not the cost!)

    Returns:
        List of same length as path containing angle cost values
        (first and last entry 0 because angle at s and t is irrelevant)
    """
    path = np.asarray(path)
    ang_out = [0]
    for p in range(len(path) - 2):
        vec1 = path[p + 1] - path[p]
        vec2 = path[p + 2] - path[p + 1]
        ang_out.append(angle(vec1, vec2))
    ang_out.append(0)
    return ang_out



import numpy as np
from numba import jit
import logging
from scipy.ndimage.morphology import distance_transform_edt

logger = logging.getLogger(__name__)


def normalize(instance):
    """
    0-1 normalization of values of instance
    """
    return (instance -
            np.min(instance)) / (np.max(instance) - np.min(instance))


@jit(nopython=True)
def rescale_instance(img, scale_factor):
    """
    Scale down image by a factor
    Arguments:
        img: numpy array of any dimension
        scale_factor: integer >= 1
    Returns:
        numpy array with 1/scale_factor size along each dimension
    """
    if scale_factor == 1:
        return img
    x_len_new = img.shape[0] // scale_factor
    y_len_new = img.shape[1] // scale_factor
    new_img = np.zeros((x_len_new, y_len_new))
    for i in range(x_len_new):
        for j in range(y_len_new):
            patch = img[i * scale_factor:(i + 1) * scale_factor,
                        j * scale_factor:(j + 1) * scale_factor]
            new_img[i, j] = np.mean(patch)
    return new_img


def rescale(instance, corridor, cfg, factor):
    """
    Prepare instance and config for the next pipeline step

    Arguments:
        instance: 2D array with resistances
        corridor: 2D array with binary (0=forbidden, 1=feasible)
        cfg: configuration with start, dest and point distances
        factor: Int > 0 : factor by which to downsample
    """
    current_cfg = cfg.copy()
    if factor > 1:
        # rescale instances
        current_instance = rescale_instance(instance, factor)
        current_corridor = (
            rescale_instance(corridor.astype(float), factor) > 0
        ).astype(int)
        # downscale start and dest
        try:
            current_cfg["start_inds"] = (np.array(cfg["start_inds"]) /
                                         factor).astype(int)
            current_cfg["dest_inds"] = (np.array(cfg["dest_inds"]) /
                                        factor).astype(int)
            # make sure start and dest are in project region
            current_corridor[tuple(current_cfg["start_inds"])] = 1
            current_corridor[tuple(current_cfg["dest_inds"])] = 1
        except KeyError:
            raise RuntimeError(
                "configuration must entail start and destination coordinates"
            )
            # downscale the KSP threshold if necessary
        if "diversity_threshold" in current_cfg.keys(
        ) and current_cfg["diversity_threshold"] > 1:
            current_cfg["diversity_threshold"
                        ] = cfg["diversity_threshold"] / factor
        # downscale the point distances
        if "point_dist_min" in current_cfg.keys():
            current_cfg["point_dist_min"] = cfg["point_dist_min"] / factor
            current_cfg["point_dist_max"] = cfg["point_dist_max"] / factor
        return current_instance, current_corridor, current_cfg
    else:
        return instance, corridor, cfg


def get_donut(radius_low, radius_high):
    """
    Compute all indices of points in donut around (0,0)
    :param radius_low: minimum radius
    :param radius_high: maximum radius
    :returns: tuples of indices of points with radius between radius_low
    and radius_high around (0, 0)
    """
    img_size = int(radius_high + 10)
    # xx and yy are 200x200 tables containing the x and y coordinates as values
    # mgrid is a mesh creation helper
    xx, yy = np.mgrid[-img_size:img_size, -img_size:img_size]
    # circle equation
    circle = (xx)**2 + (yy)**2
    # donuts contains 1's and 0's organized in a donut shape
    # you apply 2 thresholds on circle to define the shape
    donut = np.logical_and(
        circle <= (radius_high**2), circle >= (radius_low**2)
    )
    pos_x, pos_y = np.where(donut > 0)
    return pos_x - img_size, pos_y - img_size


def angle(vec1, vec2, normalize=True):
    """
    Compute angle between two vectors
    :params vec1, vec2: two 1-dim vectors of same size, can be lists or array
    :returns angle
    """
    # make array and normalize
    if normalize:
        vec1 = np.asarray(vec1)
        vec2 = np.asarray(vec2)
        vec1 = vec1 / np.linalg.norm(vec1)
        vec2 = vec2 / np.linalg.norm(vec2)
    # compute angle
    angle = np.arccos(np.clip(np.dot(vec1, vec2), -1, 1))
    # want to use full 360 degrees
    if np.sin(angle) < 0:
        angle = 2 * np.pi - angle
    # can still be nan if v1 or v2 is 0
    if np.isnan(angle):
        return 0
        # raise ValueError("angle is nan, check whether vec1 or vec2 = 0")
    return angle


def get_donut_vals(donut_tuples, vec):
    """
    compute the angle between edges defined by donut tuples
    :param donut_tuples: list of pairs of tuples, each pair defines an edge
    going from [(x1,y1), (x2,y2)]
    :param vec: vector to compute angle with
    :returns: list of same length as donut_tuples with all angles
    """
    return [angle(tup, vec) + 0.1 for tup in donut_tuples]


def get_half_donut(radius_low, radius_high, vec, angle_max=0.5 * np.pi):
    """
    Returns only the points with x >= 0 of the donut points (see above)
    :param radius_low: minimum radius
    :param radius_high: maximum radius
    :returns: tuples of indices of points with radius between radius_low
    and radius_high around (0, 0)
    """
    pos_x, pos_y = get_donut(radius_low, radius_high)
    new_tuples = []
    for i, j in zip(pos_x, pos_y):
        # compute angle
        ang = angle([i, j], vec)
        # add all valid ones
        if ang <= angle_max:
            new_tuples.append((i, j))
    return new_tuples


def compute_angle_cost(ang, max_angle_lg, mode="linear"):
    """
    Implementation of different angle cost functions:
        linear: cost increases linearly with the angle
        discrete: discrete cost classes, e.g. 0.3 for all small angles, then 1
        quadratic could be another option, just definition of a function
    Arguments:
        ang: float between 0 and pi, angle between edges
        max_angle_lg: maximum angle cutoff
        mode: "norm" or "discrete"
    returns: angle costs
    Here computed as Stefano said: up to 30 degrees + 50%, up to 60 degrees
    3 times the cost, up to 90 5 times the cost --> norm: 1.5 / 5 = 0.3
    """
    # TODO: 3 times technical costs for example
    if mode == "linear":
        return ang / max_angle_lg
    elif mode == "discrete":
        if ang == 0:
            return 0
        if ang <= np.pi / 6:
            return 0.3
        elif ang <= np.pi / 3:
            return 0.6
        elif ang <= max_angle_lg:
            return 1
        else:
            return np.inf
    elif mode == "squared":
        return (ang / max_angle_lg)**2
    else:
        raise NotImplementedError


def angle_360(vec1, vec2, normalize=True):
    if normalize:
        vec1 = np.asarray(vec1)
        vec2 = np.asarray(vec2)
        # normalize
        vec1 = vec1 / np.linalg.norm(vec1)
        vec2 = vec2 / np.linalg.norm(vec2)
    # dot product and determinant
    x1, y1 = vec1
    x2, y2 = vec2
    dot = x1 * x2 + y1 * y2  # dot product
    det = x1 * y2 - y1 * x2  # determinant
    angle = np.arctan2(det, dot)
    return angle


def get_lg_donut(
    radius_low, radius_high, vec, max_angle, max_angle_lg=np.pi / 4
):
    """
    Compute all possible combinations (tuples) of edges in restricted angle
    :param radius_low: minimum radius
    :param radius_high: maximum radius
    :param vec: direction vector
    :returns: list with entries [[edge1, edge2, cost of angle between them]]
    where costs are normalized values between 0 and 1
    """
    donut = get_donut(radius_low, radius_high)
    tuple_zip = list(zip(donut[0], donut[1]))
    linegraph_tuples = []
    for (i, j) in tuple_zip:
        # if in incoming half
        if i * vec[0] + j * vec[1] <= 0:
            for (k, l) in tuple_zip:
                ang = angle([-k, -l], [i, j])
                # if smaller max angle and general outgoing half
                if ang <= max_angle_lg and k * vec[0] + l * vec[1] >= 0:
                    angle_norm = compute_angle_cost(ang, max_angle_lg)
                    linegraph_tuples.append([[i, j], [k, l], angle_norm])
    return linegraph_tuples


def get_path_lines(cost_shape, paths):
    """
    Given a list of paths, compute continous lines in an array of cost_shape
    :param cost_shape: desired 2-dim output shape of array
    :param paths: list of paths of possibly different lengths, each path is
    a list of tuples
    :returns: 2-dim binary of shape cost_shape where paths are set to 1
    """
    path_dilation = np.zeros(cost_shape)
    for path in paths:
        # iterate over path nodes
        for i in range(len(path) - 1):
            line = bresenham_line(*path[i], *path[i + 1])
            # set all pixels on line to 1
            for (j, k) in line:
                path_dilation[j, k] = 1
    return path_dilation


def get_pipeline(num_vertices, num_shifts, mem_limit):
    """
    Compute the optimal downsampling factors in an iterative pipeline approach
    Arguments:
        num_vertices: Int, number of vertices in the graph
        num_shifts: Int, number of neighbors for each vertex
        mem_limit: Int, maximal number of edges allowed
    """
    factor = 1
    # return first one that fits in mem_limit
    while (num_vertices * num_shifts) / factor**4 > mem_limit:
        factor += 1
    return np.arange(factor, 0, -1)


def pipeline_corridor(paths, out_shape, n_shifts, mem_limit, next_factor):
    """
    Get the next corridor in a pipeline (automatically calibrate the corridor
    width based on the next sampling factor)

    Arguments:
        path_points: list of arrays of shape (n, 2) containing the points on one
                or more paths that have been found in the previous iteration
        out_shape: Tuple, shape of array that will be the output corridor
        n_shifts: Int, Number of neighbors per vertex (based on point_dist_min
                and point_dist_max)
        mem_limit: Int, Maximum number of edges
        next_factor: Int, downsampling factor in the upcoming next iteration
    """
    # close gap between points on path with bresenham line
    distance_transform = np.ones(out_shape)
    for path_points in paths:
        for i in range(len(path_points) - 1):
            line = bresenham_line(*path_points[i], *path_points[i + 1])
            for (x, y) in line:
                distance_transform[x, y] = 0
    # compute distance transform
    # distance_transform[x,y] = min distance of cell (x,y) to the path(s)
    distance_transform = distance_transform_edt(distance_transform)
    # check how many pixels you get when setting the corridor to all cells with
    # distance less than 20 (in order to test how many cells you get then)
    test_corridor_width = 20
    test_corridor = (distance_transform < test_corridor_width).astype(int)
    # Estimate the number of edges that one will get with the 20-cell corridor:
    # #edges = #pixels_in_corridor x #neighbors / factor**4 because a lower
    # resolution reduces #pixels quadratically and #neighbors quadtratically
    estimated_edges_20 = (np.sum(test_corridor) * n_shifts) / (next_factor**4)
    # (mem_limit / estimated_edges_20) is then the factor how much more or less
    # cells we can include in the corridor. E.g. With 20 we get 100 edges, but
    # mem_limit = 1000 --> we can use 200 instead of 20 as the corridor width
    # (but must be at least 10)
    optimal_corridor_width = max(
        [test_corridor_width * mem_limit / estimated_edges_20, 3]
    )
    # get the corridor with all cells closer than optimal_corridor_width
    corridor = (distance_transform <= optimal_corridor_width).astype(int)
    logger.info(
        f"Next corridor around path was set to width {optimal_corridor_width}"
    )
    return corridor


def bresenham_line(x0, y0, x1, y1):
    """
    Finds the cell indices on a straight line between two raster cells
    """
    steep = abs(y1 - y0) > abs(x1 - x0)
    if steep:
        x0, y0 = y0, x0
        x1, y1 = y1, x1

    switched = False
    if x0 > x1:
        switched = True
        x0, x1 = x1, x0
        y0, y1 = y1, y0

    if y0 < y1:
        ystep = 1
    else:
        ystep = -1

    deltax = x1 - x0
    deltay = abs(y1 - y0)
    error = -deltax / 2
    y = y0

    line = []
    for x in range(x0, x1 + 1):
        if steep:
            line.append([y, x])
        else:
            line.append([x, y])

        error = error + deltay
        if error > 0:
            y = y + ystep
            error = error - deltax
    if switched:
        line.reverse()
    return line



import numpy as np
import pickle
import time
import os
import argparse
import matplotlib.pyplot as plt

from lion.utils import general
from lion.algorithms import (
    optimal_point_spotting, optimal_route, ksp_points, ksp_routes
)


def plot_paths(plot_instance, paths, buffer=0, out_path="test_path.png"):

    # normalize instance
    plotmin = np.min(plot_instance[plot_instance < np.inf])
    plotmax = np.max(plot_instance[plot_instance < np.inf])
    instance = (plot_instance - plotmin) / (plotmax - plotmin)

    expanded = np.expand_dims(instance, axis=2)
    expanded = np.tile(expanded, (1, 1, 3))  # overwrite instance by tiled one
    # colour nodes in path in red
    for path in paths:
        for (x, y) in path:
            expanded[x - buffer:x + buffer + 1, y - buffer:y + buffer +
                     1] = [0.9, 0.2, 0.2]  # colour red
    # plot and save
    plt.figure(figsize=(25, 25))
    plt.imshow(np.swapaxes(expanded, 1, 0))
    if out_path is not None:
        plt.savefig(out_path, bbox_inches='tight')
    else:
        plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--path", help="Path to pickled example data")
    parser.add_argument(
        "-s",
        "--save_path",
        default=None,
        type=str,
        help="Path where to save an output image and array of\
             the path if desired"
    )
    args = parser.parse_args()

    # load instance
    with open(args.path, "rb") as infile:
        data = pickle.load(infile)

    # get the array and config from the loaded data
    try:
        test_instance = data["matrix"]
        cfg = data["configs"][0]
    except (KeyError, ValueError) as e:
        raise e(
            "Wrong format of pickle file: must be a dictionary\
            with a key 'matrix' and a key 'configs' where matrix is a numpy\
            array and configs is a list with one dictionary inside"
        )

    print("shape of instance and cfg:", test_instance.shape, cfg)
    print("unique values in instance", np.unique(test_instance))

    # modify point distance (currently not provided)
    if "point_dist_min" not in cfg.keys():
        cfg["point_dist_min"] = 15
        cfg["point_dist_max"] = 25

    tic = time.time()
    path = optimal_point_spotting(test_instance, cfg.copy())
    paths = [path]
    # paths = ksp_pylons(test_instance, cfg.copy(), 5)
    print("overall time for processing", time.time() - tic)

    # save outputs
    if args.save_path is not None:
        # plotting - set high values to inf because otherwise everthing except
        # for the high cost area will be just black
        plot_instance = test_instance.copy().astype(float)
        plot_instance[plot_instance == np.max(plot_instance)] = np.inf
        # set buffer: Bigger instance --> use bigger buffer
        plot_paths(
            plot_instance,
            paths,
            buffer=2,
            out_path=os.path.join(
                os.path.join(args.save_path, "test_output.png")
            )
        )
        with open(os.path.join(args.save_path, "paths.dat"), "wb") as outfile:
            pickle.dump(paths, outfile)



"""Package installer."""
import os
from setuptools import setup
from setuptools import find_packages

LONG_DESCRIPTION = ''
if os.path.exists('README.md'):
    with open('README.md') as fp:
        LONG_DESCRIPTION = fp.read()

scripts = []

setup(
    name='lion-sp',
    version='0.1.0',
    description='LInear Optimization Networks - shortest path algorithms',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Nina Wiedemann',
    author_email=('nwiedemann@uos.de'),
    license="GPLv3",
    url='https://github.com/NinaWie/lion',
    install_requires=['numpy', 'numba', 'scipy', 'coverage'],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    packages=find_packages('.'),
    python_requires='>=3.6',
    scripts=scripts
)



from numba import jit
import numpy as np
from lion.utils.general import angle, angle_360


def get_update_algorithm(func, angle_cost_array, in_edges, out_edges=None):
    """
    Interface to update algorithms - dependent on the angle cost function,
    choose the appropriate update algorithm and return it with the necessary
    arguments

    Arguments:
        func: str defining the angle cost function
        angle_cost_array: n x n 2D numpy array containing the costs for each
            combination of vectors
        in_edges: n x 2 shaped array with vectors of incoming edges
        out_edges: n x 2 shaped array with vectors of outgoing edges
    Returns:
        algorithm: function, update algorithm to be used
        args: Tuple, corresponding arguments
    """
    if func == "linear" and len(angle_cost_array) > 50:
        # >50 because hidden factors in O-notation - will be slower
        args = (angle_cost_array)
        algorithm = update_linear
    elif func == "discrete" and len(angle_cost_array) > 100:
        # >100 because hidden factors in O-notation - will be slower
        args = prepare_for_discrete(angle_cost_array, in_edges, out_edges)
        algorithm = update_discrete
    else:
        args = (angle_cost_array)
        algorithm = update_default
    return algorithm, args


def prepare_for_discrete(angle_cost_array, in_edges, out_edges=None):
    """
    Computation of special arguments for update algorithm for discrete angle
    cost functions
    Arguments:
        angle_cost_array: n x n 2D numpy array containing the costs for each
            combination of vectors
        in_edges: n x 2 shaped array with vectors of incoming edges
        out_edges: n x 2 shaped array with vectors of outgoing edges
    Returns:
        alphas: Angles of incoming edges with respect to reference vector
        tree_index: 1D array, sorted indices of outgoing angles
        tree_values: 1D array, sorted values of outgoing angles
        discrete_costs: 1D array of length d, possible angle costs (discrete
                function, so should only be a few)
        bounds: 2D array of shape d x 2, minimum and maximum angle for each
                possible angle cost
    """
    # normalize in and out edges:
    in_edges_norm = [np.asarray(s) for s in in_edges]
    in_edges_norm = [s / np.linalg.norm(s) for s in in_edges_norm]
    if out_edges is not None:
        out_edges_norm = [np.asarray(s) for s in out_edges]
        out_edges_norm = [s / np.linalg.norm(s) for s in out_edges_norm]
    else:
        out_edges_norm = in_edges_norm

    # get unique discrete costs and bounds
    discrete_costs = np.unique(angle_cost_array)
    d_uni = len(discrete_costs)

    # get bounds
    angles_raw = np.array(
        [
            [
                angle(in_edge, out_edge, normalize=False)
                for in_edge in in_edges_norm
            ] for out_edge in out_edges_norm
        ]
    )
    bounds = np.zeros((d_uni, 2))
    prev_max = 0
    for i, dis_cost in enumerate(discrete_costs):
        covered_angles = angles_raw[angle_cost_array == dis_cost]
        in_between = 0.5 * (prev_max + np.min(covered_angles))
        if i > 0:
            bounds[i - 1, 1] = in_between
        prev_max = np.max(covered_angles)
        bounds[i, 0] = in_between
    bounds[i, 1] = prev_max + 1

    # compute alphas and betas
    ref_edge = [1, 0]
    alphas = np.asarray(
        [
            angle_360(e, ref_edge, normalize=False) + np.pi
            for e in in_edges_norm
        ]
    )
    betas = np.array(
        [
            angle_360(e, ref_edge, normalize=False) + np.pi
            for e in out_edges_norm
        ]
    )

    tree_index = np.argsort(betas)
    tree_values = betas[tree_index]

    return alphas, tree_index, tree_values, discrete_costs, bounds


@jit(nopython=True)
def update_default(dists, angles_all):
    """
    Trivial updates working for any angle cost function
    O(kl) for k in edges and l out edges
    """
    predecessors = np.zeros(len(angles_all), dtype=np.int64)
    for s in range(len(angles_all)):
        in_costs = dists + angles_all[s]
        predecessors[s] = np.argmin(in_costs)
    return predecessors


@jit(nopython=True)
def update_linear(dists, angles_all):
    """
    Efficient update mechanism for LINEAR angle cost functions, i.e. the angle
    cost increases linearly with the angle
    Improves the runtime compared to update_default to O((k+l) log kl) instead
    of O(kl) for a vertex update with k incoming and l outgoing edges

    Arguments:
        angles_all: 2Darray of size (len(shifts), len(shifts))
                    precomputed angle cost for each tuple of edges
        dists: 1D array of size (len(shifts)) which are the distances to the
                    incoming edges
    Returns:
        1D int array containing the optimal predecessor for each out edge
    """
    n_neighbors = len(angles_all)

    # sort the in edge distances and initialize
    initial_S = np.argsort(dists)
    marked_plus = np.zeros(n_neighbors)
    marked_minus = np.zeros(n_neighbors)

    # for now, assign all edges its straight line predecessor
    predecessors = np.arange(n_neighbors)
    preliminary_dist = np.zeros(n_neighbors)
    for s in range(n_neighbors):
        preliminary_dist[s] = dists[s] + angles_all[s, s]

        # set current tuple: in edge and shift
    # (out edge index unncessary because same as in edge)
    current_in_edge = initial_S[0]
    current_shift = 0
    tuple_counter = 0

    while tuple_counter < len(initial_S) - 1:
        # best out edge is exactly the same shift!
        current_out_edge = (current_in_edge + current_shift) % n_neighbors

        # compute possible update value:
        update_val = dists[current_in_edge] + angles_all[current_out_edge,
                                                         current_in_edge]

        if current_shift == 0:
            marked = marked_plus[current_out_edge] and marked_minus[
                current_out_edge]
        elif current_shift > 0:
            marked = marked_plus[current_out_edge]
        else:
            marked = marked_minus[current_out_edge]

        # update only if better
        if marked == 0 and np.around(update_val, 5) <= np.around(
            preliminary_dist[current_out_edge], 5
        ):
            preliminary_dist[current_out_edge] = update_val
            predecessors[current_out_edge] = int(current_in_edge)

            # progress one edge further
            progress_one = True

        # already marked or update not successful:
        # Consider first edge in other direction or next overall tuple
        else:
            progress_one = False
            if current_shift > 0:
                current_shift = -1
            else:
                # get next tuple from stack
                tuple_counter += 1
                current_in_edge = initial_S[tuple_counter]
                current_shift = 0

        # Progress to next edge
        if progress_one:
            if current_shift < 0:
                current_shift -= 1
            if current_shift <= 0:
                marked_minus[current_out_edge] = 1
            if current_shift >= 0:
                current_shift += 1
                marked_plus[current_out_edge] = 1

    return predecessors


@jit(nopython=True)
def _circular_plus(a, b):
    """
    Plus op on circular x-axis: If sum is greater than 2pi, rotate to beginning
    Arguments:
        a,b: floats between 0 and 2 * np.pi (angle values)
    """
    return (a + b) % (2 * np.pi)


@jit(nopython=True)
def _binary_index_search(key, tree_values, tree_index):
    """
    Binary search to find the closest (lower) value to <key> in an array
    (recursive method)

    Arguments:
        key: Float, the value to look for
        tree_values: 1D sorted array of values in the tree
        tree_index: 1D int array of same length as tree_values, contains the
                original indices for the sorted value of tree_values
    Returns:
        The index in tree_index where the corresponding value in tree_values
        is closest to key (closest from left side, i.e. the lower one)
    """
    tree_len = len(tree_values)
    if tree_len == 1:
        return tree_index[0]
    middle = tree_len // 2
    if key < tree_values[middle]:
        return _binary_index_search(
            key, tree_values[:middle], tree_index[:middle]
        )
    else:
        return _binary_index_search(
            key, tree_values[middle:], tree_index[middle:]
        )


@jit(nopython=True)
def _circular_minus(a, b):
    """
    Minus op on circular x-axis - if smaller zero, then subtract from 2 pi
    Arguments:
        a,b: floats between 0 and 2 * np.pi (angle values)
    """
    if a < b:
        return 2 * np.pi - b + a
    else:
        return a - b


@jit(nopython=True)
def update_discrete(dists, args):
    """
    Efficient update mechanism for DISCRETE angle cost functions, i.e. there
    are only d possible angle costs
    Improves the runtime compared to update_default to O(dk log l) instead
    of O(kl) for a vertex update with k incoming and l outgoing edges

    Arguments:
        angles_all: 2Darray of size (len(shifts), len(shifts))
                    precomputed angle cost for each tuple of edges
        dists: 1D array of size (len(shifts)) which are the distances to the
                    incoming edges
    Returns:
        1D int array containing the optimal predecessor for each out edge
    """
    (alphas, tree_index, tree_values, discrete_costs, bounds) = args
    # sort tuples of dists and possible angle costs
    d_uni = len(discrete_costs)
    n_in_edges = len(alphas)

    dists_tuples = np.zeros(n_in_edges * d_uni)
    for i in range(n_in_edges):
        for d in range(d_uni):
            dists_tuples[d * n_in_edges + i] = dists[i] + discrete_costs[d]
    sorted_inds = np.argsort(dists_tuples)
    sorted_tuples = np.zeros((len(sorted_inds), 2), dtype=np.int64)
    for s in range(len(sorted_inds)):
        ind = sorted_inds[s]
        sorted_tuples[s, 0] = ind % n_in_edges
        sorted_tuples[s, 1] = ind // n_in_edges

    predecessor = np.zeros(len(tree_values)) - 1
    is_updated = 0

    for s in range(len(sorted_tuples)):
        in_edge = sorted_tuples[s, 0]
        dis_step = sorted_tuples[s, 1]
        # in edge is the index of the incoming edge --> get corresponding alpha
        for j in range(2):
            # compute the current range of betas
            if j == 0:
                min_angle_bound = _circular_minus(
                    alphas[in_edge], bounds[dis_step, 1]
                )
                max_angle_bound = _circular_minus(
                    alphas[in_edge], bounds[dis_step, 0]
                )
            else:
                min_angle_bound = _circular_plus(
                    alphas[in_edge], bounds[dis_step, 0]
                )
                max_angle_bound = _circular_plus(
                    alphas[in_edge], bounds[dis_step, 1]
                )
            # find the corresponding range of indices (binary search)
            fake_ind = np.arange(len(tree_values))
            min_angle_index = _binary_index_search(
                min_angle_bound, tree_values, fake_ind
            )
            if min_angle_index != 0 or min_angle_bound > tree_values[0]:
                min_angle_index += 1
            max_angle_index = _binary_index_search(
                max_angle_bound, tree_values, fake_ind
            ) + 1
            if max_angle_bound < min_angle_bound:
                inds_inbetween = np.concatenate(
                    (
                        tree_index[min_angle_index:],
                        tree_index[:max_angle_index]
                    )
                )
            else:
                inds_inbetween = tree_index[min_angle_index:max_angle_index]
            # update predecessor
            for index in inds_inbetween:
                if predecessor[index] < 0:
                    predecessor[index] = in_edge
                    is_updated += 1
        # TODO: experiment whether it's worth to make the tree smaller
        if is_updated == len(tree_values):
            break
    return predecessor



import numpy as np
import unittest
from types import SimpleNamespace
from lion.angle_graph import AngleGraph
from lion.utils.general import bresenham_line
from lion.utils.costs import compute_geometric_costs


class TestImplicitLG(unittest.TestCase):

    expl_shape = (50, 50)

    # create configuration
    cfg = SimpleNamespace()
    cfg.point_dist_min = 3
    cfg.point_dist_max = 5
    start_inds = np.array([6, 6])
    dest_inds = np.array([41, 43])
    cfg.start_inds = start_inds
    cfg.dest_inds = dest_inds
    cfg.angle_weight = 0.25
    cfg.edge_weight = 0
    cfg.max_angle = np.pi / 2
    cfg.max_angle_lg = np.pi / 4
    cfg.layer_classes = ["dummy_class"]
    cfg.class_weights = [1]

    # construct simple line instance
    example_inst = np.ones(expl_shape)
    # construct corresponding corridor
    working_expl_corr = np.zeros(expl_shape)
    line = bresenham_line(
        start_inds[0], start_inds[1], dest_inds[0], dest_inds[1]
    )
    for (i, j) in line:
        working_expl_corr[i - 1:i + 1, j - 1:j + 1] = 1

    # construct instance that required 90 degree angle
    high_angle_corr = np.zeros(expl_shape)
    high_angle_corr[start_inds[0], start_inds[1]:dest_inds[1] - 4] = 1
    high_angle_corr[start_inds[0] + 4:dest_inds[0] + 1, dest_inds[1]] = 1
    # no_connection_corr: there is no way to get from start to dest at all
    # with given point_dist_max
    no_connection_corr = high_angle_corr.copy()
    # high_angle_corr: there is a connection, but it requires a sharp turn
    # and is not possible with smaller max_angle_lg
    high_angle_corr[start_inds[0], dest_inds[1]] = 1

    def test_correct_shortest_path(self) -> None:
        """ Test the implicit line graph construction """
        graph = AngleGraph(self.example_inst, self.working_expl_corr)
        path = graph.single_sp(**vars(self.cfg))
        path, path_costs, cost_sum = graph.transform_path(path)
        self.assertTupleEqual(graph.instance.shape, self.expl_shape)
        self.assertEqual(graph.angle_weight + graph.resistance_weight, 1)
        self.assertNotEqual(len(graph.shifts), 0)
        # all initialized to infinity
        # self.assertTrue(not np.any([graph.dists[graph.dists < np.inf]]))
        # start point was set to normalized value
        start_ind = graph.pos2node[tuple(self.start_inds)]
        self.assertEqual(0, graph.dists[start_ind, 0])
        # all start dists have same value
        self.assertEqual(len(np.unique(graph.dists[start_ind])), 1)
        # not all values still inf
        self.assertLess(np.min(graph.dists[start_ind]), np.inf)
        # get actual best path
        # path, path_costs, cost_sum = graph.get_shortest_path(
        #     self.start_inds, self.dest_inds
        # )
        self.assertNotEqual(len(path), 0)
        self.assertEqual(len(path), len(path_costs))
        self.assertGreaterEqual(cost_sum, 5)
        for (i, j) in path:
            self.assertEqual(self.example_inst[i, j], 1)

    def test_edge_costs(self) -> None:
        graph = AngleGraph(self.example_inst, self.working_expl_corr)
        self.cfg.angle_weight = 0
        self.cfg.edge_weight = 0.5
        path = graph.single_sp(**vars(self.cfg))
        path, _, cost_sum = graph.transform_path(path)
        dest_ind = graph.pos2node[tuple(self.dest_inds)]
        dest_costs = np.min(graph.dists[dest_ind])
        a = []
        path = np.array(path)
        weighted_inst = self.example_inst
        for p in range(len(path) - 1):
            line = bresenham_line(
                path[p, 0], path[p, 1], path[p + 1, 0], path[p + 1, 1]
            )[1:-1]
            line_costs = [weighted_inst[i, j] for (i, j) in line]
            line_costs = [li for li in line_costs if li < np.inf]
            bresenham_edge_dist = np.mean(line_costs) * self.cfg.edge_weight
            shift_costs = np.linalg.norm(path[p] - path[p + 1])
            # append edge cost
            a.append(
                shift_costs * (
                    0.5 * (
                        weighted_inst[tuple(path[p])] +
                        weighted_inst[tuple(path[p + 1])]
                    ) + bresenham_edge_dist
                )
            )
        dest_costs_gt = np.sum(a)
        self.assertTrue(np.isclose(cost_sum, dest_costs_gt))
        self.assertTrue(np.isclose(dest_costs, dest_costs_gt))
        out_compute_cost_method = compute_geometric_costs(
            path, weighted_inst, edge_weight=self.cfg.edge_weight
        )
        self.assertEqual(np.sum(out_compute_cost_method), cost_sum)

        self.cfg.angle_weight = 0.25
        self.cfg.edge_weight = 0

    def test_angle_sp(self) -> None:
        graph = AngleGraph(self.example_inst, self.high_angle_corr)
        _ = graph.single_sp(**vars(self.cfg))
        # assert that destination can NOT be reached
        dest_ind = graph.pos2node[tuple(self.dest_inds)]
        self.assertFalse(np.min(graph.dists[dest_ind]) < np.inf)
        # Note: this test does not change the stack array because the path is
        # only empty because of an inf in angles_all
        # NEXT TRY: more angles allowed
        self.cfg.max_angle_lg = np.pi
        graph = AngleGraph(self.example_inst, self.high_angle_corr)
        path = graph.single_sp(**vars(self.cfg))
        path, _, cost_sum = graph.transform_path(path)
        # assert that dest CAN be reached
        dest_ind = graph.pos2node[tuple(self.dest_inds)]
        self.assertTrue(np.min(graph.dists[dest_ind]) < np.inf)

    def test_empty_path(self) -> None:
        self.cfg.max_angle_lg = np.pi
        graph = AngleGraph(self.example_inst, self.no_connection_corr)
        _ = graph.single_sp(**vars(self.cfg))
        # assert that destination can NOT be reached
        dest_ind = graph.pos2node[tuple(self.dest_inds)]
        self.assertFalse(np.min(graph.dists[dest_ind]) < np.inf)


if __name__ == '__main__':
    unittest.main()



from numba import jit
import numpy as np


@jit(nopython=True)
def topological_sort_jit(v_x, v_y, shifts, to_visit, stack, curr_ind):
    """
    Recursive method for topological sorting of vertices in a DAG

    Arguments:
        v_x, v_y: current vertex / cell
        shifts: array of length n_neighborsx2 to iterate over neighbors
        to_visit: 2D array of size of instance to remember visited nodes
        stack: list of topologically sorted vertices
    Returns:
        stack of sorted cells
    """
    x_len, y_len = to_visit.shape
    # Mark the current node as visited.
    to_visit[v_x, v_y] = 0
    # Recur for all the vertices adjacent to this vertex
    for s in range(len(shifts)):
        neigh_x = v_x + shifts[s, 0]
        neigh_y = v_y + shifts[s, 1]
        # in array bounds and not visited yet
        if (
            neigh_x >= 0 and neigh_x < x_len and neigh_y >= 0
            and neigh_y < y_len and to_visit[neigh_x, neigh_y] == 1
        ):
            _, curr_ind = topological_sort_jit(
                neigh_x, neigh_y, shifts, to_visit, stack, curr_ind
            )
    # Push current vertex to stack which stores result
    stack[v_x, v_y] = curr_ind
    return stack, curr_ind + 1


@jit(nopython=True)
def edge_costs(
    stack, pos2node, shifts, edge_cost, instance, edge_inst, shift_lines,
    shift_costs, edge_weight
):
    """
    Pre-compute all costs on each edge from resistances at the points or if
    edge_weight>0 then also including resistances between points

    Arguments:
        stack: np array of shape (n,2): order in which to consider the vertices
            MUST BE TOPOLOGICALLY SORTED for this algorithm to work
        pos2node: 2D ndarray with pos2node[x,y] = index of cell (x,y) in stack
        shifts: np array of size (x,2), indicating the neighborhood for each
            vertex
        edge_cost: 2Darray of size (n, len(shifts)), initially all inf
        instance: 2Darray of point resistance values for each cell
        edge_inst: 2Darray with resistances to traverse this cell
                   (often same as instance)
        shift_lines: numba typed List filled with len(shifts) np arrays,
                    each array of shape (x,2) is the Bresenham line connecting
                    a cell to one of its neighbors
        shift_costs: 1Darray of length len(shift) containing the Euclidean
                    length to each neighbor
        edge_weight: Weight defining importance of costs between points
                compared to costs at the points themselves
    """
    edge_inst_len_x, edge_inst_len_y = edge_inst.shape
    # print(len(stack))
    for i in range(len(stack)):
        v_x = stack[i, 0]
        v_y = stack[i, 1]
        vertex_costs = instance[v_x, v_y]
        for s in range(len(shifts)):
            neigh_x = v_x + shifts[s][0]
            neigh_y = v_y + shifts[s][1]
            if (
                0 <= neigh_x < edge_inst_len_x
                and 0 <= neigh_y < edge_inst_len_y
                and pos2node[neigh_x, neigh_y] >= 0
            ):
                bresenham_edge_dist = 0
                if edge_weight > 0:
                    bres_line = shift_lines[s] + np.array([v_x, v_y])
                    edge_cost_list = np.zeros(len(bres_line) - 2)
                    for k in range(1, len(bres_line) - 1):
                        edge_cost_list[k - 1] = edge_inst[bres_line[k][0],
                                                          bres_line[k][1]]
                    # TODO: mean or sum?
                    bresenham_edge_dist = edge_weight * np.mean(edge_cost_list)
                neigh_ind = pos2node[neigh_x, neigh_y]
                edge_cost[neigh_ind, s] = shift_costs[s] * (
                    0.5 * (vertex_costs + instance[neigh_x, neigh_y]) +
                    bresenham_edge_dist
                )
    return edge_cost


@jit(nopython=True)
def sp_dag(
    stack, pos2node, shifts, angles_all, dists, edge_cost, algorithm, *args
):
    """
    Angle-weighted dynamic program for Directed Acyclic Graphs (O(n))
    Stores the distances and predecessors for all IN edges for each vertex

    Arguments:
        stack: np array of shape (n,2): order in which to consider the vertices
            MUST BE TOPOLOGICALLY SORTED for this algorithm to work
        pos2node: 2D ndarray with pos2node[x,y] = index of cell (x,y) in stack
        shifts: np array of size (x,2) --> indicating the neighborhood for each
            vertex
        angles_all: 2Darray of size (len(shifts), len(shifts))
                    precomputed angle cost for each tuple of edges
        dists: 2Darray of size (n, len(shifts)) - contains distance of each
               edge from the source vertex
        edge_cost: 2Darray of size (n, len(shifts)) - edge cost for each edge
    """
    inst_x_len, inst_y_len = pos2node.shape
    preds = np.zeros(dists.shape) - 1
    # print(len(stack))
    for i in range(len(dists)):
        v_x = stack[i, 0]
        v_y = stack[i, 1]

        # get predecessor for all outgoing edges at this vertex
        predecessors = algorithm(dists[i], *args)

        # get index and update
        for s in range(len(shifts)):
            neigh_x = int(v_x + shifts[s][0])
            neigh_y = int(v_y + shifts[s][1])
            if (
                0 <= neigh_x < inst_x_len and 0 <= neigh_y < inst_y_len
                and pos2node[neigh_x, neigh_y] >= 0
            ):
                neigh_stack_ind = pos2node[neigh_x, neigh_y]
                # add up point cost + angle cost + edge cost
                pred = int(predecessors[s])
                cost_and_angle = dists[i, pred] + angles_all[s, pred]

                # update distances and predecessors
                dists[neigh_stack_ind,
                      s] = cost_and_angle + edge_cost[neigh_stack_ind, s]
                preds[neigh_stack_ind, s] = pred
    return dists, preds


@jit(nopython=True)
def sp_dag_reversed(
    stack, pos2node, shifts, angles_all, dists, edge_cost, algorithm, *args
):
    """
    Angle-weighted dynamic program for Directed Acyclic Graphs (O(n))
    Stores the distances and predecessors for all OUT edges for each vertex

    Arguments:
        stack: np array of shape (n,2): order in which to consider the vertices
            MUST BE TOPOLOGICALLY SORTED for this algorithm to work
        pos2node: 2D ndarray with pos2node[x,y] = index of cell (x,y) in stack
        shifts: np array of size (x,2) --> indicating the neighborhood for each
            vertex
        angles_all: 2Darray of size (len(shifts), len(shifts))
                    precomputed angle cost for each tuple of edges
        dists: 2Darray of size (n, len(shifts)) - contains distance of each
               edge from the source vertex
        preds: 2Darray of size (n, len(shifts)) - contains optimal predecessor
               of each edge from the source vertex
        edge_cost: 2Darray of size (n, len(shifts)) - edge cost for each edge
    """

    inst_len_x, inst_len_y = pos2node.shape
    n_neighbors = len(shifts)
    preds = np.zeros(dists.shape) - 1

    for i in range(len(stack)):
        v_x = stack[-i - 1, 0]
        v_y = stack[-i - 1, 1]

        # get in distances by computing the in neighbors
        in_inds = np.zeros(n_neighbors) - 1
        in_dists = np.zeros(n_neighbors) + np.inf
        for s_in in range(n_neighbors):
            in_neigh_x = v_x - shifts[s_in][0]
            in_neigh_y = v_y - shifts[s_in][1]
            if (
                0 <= in_neigh_x < inst_len_x and 0 <= in_neigh_y < inst_len_y
                and pos2node[in_neigh_x, in_neigh_y] >= 0
            ):
                in_neigh_ind = pos2node[in_neigh_x, in_neigh_y]
                in_inds[s_in] = int(in_neigh_ind)
                in_dists[s_in] = dists[in_neigh_ind, s_in]

        predecessors = algorithm(in_dists, *args)

        # update OUTGOING edge distances
        for s in range(n_neighbors):
            neigh_x = v_x + shifts[s][0]
            neigh_y = v_y + shifts[s][1]

            # out neighbor feasible?
            if (0 <= neigh_x <
                inst_len_x) and (0 <= neigh_y < inst_len_y
                                 ) and (pos2node[neigh_x, neigh_y] >= 0):

                # beginning: only update out edges of destination
                if i == 0:
                    dists[-1, s] = edge_cost[-i - 1, s]
                    continue
                pred = int(predecessors[s])
                cost_and_angle = in_dists[pred] + angles_all[s, pred]
                dists[-i - 1, s] = cost_and_angle + edge_cost[-i - 1, s]
                preds[-i - 1, s] = pred
    return dists, preds
